/*
 * gui-gtk.c - the gui for createrec-gtk
 * Copyright (c) 2009 Markus Kohm
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "gui.h"
#include "ifile.h"
#include <sys/wait.h>
#include <glib/gi18n.h>
#include <stdlib.h>

struct _gui_context {
    gdouble       progress;     /* Last progress */
    gchar         *inputvideo;  /* video file to use */
    gchar         *tmpdir;
    gchar         *outputdir;
    gchar         *title;
    guint         number;
    gboolean      searchfree;
    IFileContext  *iF_context;
};

extern gboolean createrec_do( const gchar *dirname, 
			      guint number, gboolean searchfirstfree,
			      IFileContext *iF_context, const gchar *tmpdir,
			      GuiContext *gui_context, 
			      GError **err );

/**
 * Show a GError message
 */
static void gui_error_message( GuiContext *gui_context, GError **err ) {
    g_assert( gui_context );
    g_assert( err );
    g_assert( *err );

    g_error( (*err)->message );
    
    g_clear_error( err ); /* We will never be here, because g_error exits */
}

/**
 * Show progress or state of the programm
 *
 * @param progress < 0 for not changing the progress, 
 *                 0.0..1.0 for changing progress.
 * @param message is a state string or NULL if state not changed
 * @param gui_context pointer to a gui context.
 */
void gui_show_state( gdouble progress, gchar *message,
		     GuiContext *gui_context ) {
    g_assert( gui_context );
    if ( message ) {
	g_print( "%-79s\n", message );
    }
    if ( !(progress < 0.0) && !(progress > 1.0) ) {
	if ( ( gui_context->progress == 0.0 && progress > 0.0 )
	     || progress == 1.0
	     || progress < gui_context->progress
	     || progress - gui_context->progress > 0.02 ) {
	    gchar s_bar[51];
	    guint n;
	    guint n_is = (guint)(progress * 50);
	    for ( n = 0; n < n_is; n++ )
		if ( ( n % 5 ) == 4 )
		    s_bar[n] = '#';
		else
		    s_bar[n] = '+';
	    for ( ; n < 50; n++ )
		if ( ( n % 5 ) == 4 )
		    s_bar[n] = '|';
		else
		    s_bar[n] = '.';
	    s_bar[50] = 0;
	    g_print( "%s\r", s_bar );
	    gui_context->progress = progress;
	}
    }
    if ( progress >= 1.0 )
	g_print( "\n" );
}

static void glf_dummy(const gchar *log_domain,
		      GLogLevelFlags log_level,
		      const gchar *message,
		      gpointer user_data) {
}

/**
 */
static void gui_show_version( void ) {
    g_print( "createrec_cli %s\n", VERSION );
    g_print( "Copyright (c) 2009 Markus Kohm\n\n" );
    g_print( "This program comes with ABSOLUTELY NO WARRANTY.\n" );
    g_print( "This is free software, and you are welcome\n" );
    g_print( "to redistribute it under certain conditions.\n" );
    g_print( "See COPYING for details.\n" );
    exit(0);
}

/**
 * Do all initialization for the gui without entering the main loop.
 *
 * Note, that argc and argv are input and output parameters.  After
 * the initialization all gui parameters are removed from the array.
 * You should call this only once!
 * 
 * @param argc pointer to number of arguments
 * @param argv array of arguments
 * @param gui_context pointer to a gui gui_context.
 * @return TRUE if ok, FALSE on error
 */
gboolean gui_init(int *argc, char ***argv, GuiContext **gui_context) {
    static gboolean onceonly = FALSE;
    GError *err = NULL;
    GuiContext *gc;
    
    g_assert( argc != NULL && argv != NULL && !onceonly );
    g_assert( gui_context != NULL );
    gc = g_malloc0( sizeof( GuiContext ) );
    onceonly = TRUE;

    if ( !ifile_init(argc, argv, &(gc->iF_context), &err) )
	g_error( "ToDO: add handler for ifile_init: %s\n", err->message );
    else {
	GError *error = NULL;
	GOptionContext *context;
	gboolean bdebug = FALSE;
	GOptionEntry entries[] = {
	    { "inputvideo", 'i', 0, G_OPTION_ARG_STRING, &gc->inputvideo, N_("video file to create record from"), N_("filename") },
	    { "tmpdir", 't', 0, G_OPTION_ARG_STRING, &gc->tmpdir, N_("directory for temporary files"), N_("directoryname") },
	    { "outputdir", 'o', 0, G_OPTION_ARG_STRING, &gc->outputdir, N_("directory for records"), "directoryname" },
	    { "outputnumber", 'n', 0, G_OPTION_ARG_INT, &gc->number, N_("number of record"), "0..9999" },
	    { "searchfree", 's', 0, G_OPTION_ARG_NONE, &gc->searchfree, N_("search for first unused record number"), NULL },
	    { "title", 'T', 0, G_OPTION_ARG_STRING, &gc->title, N_("title used for the record"), N_("string") },
	    { "debug", 0, 0, G_OPTION_ARG_NONE, &bdebug, N_("activate debugging output"), NULL },
	    { "version", 0, G_OPTION_FLAG_NO_ARG, G_OPTION_ARG_CALLBACK, gui_show_version, N_("show version information only"), NULL }
	};
	context = g_option_context_new( _(" - generate record from video") );
	g_option_context_add_main_entries(context, entries, NULL);
	if ( !g_option_context_parse(context,argc,argv,&err) ) {
	    g_error( _("option parsing failed: %s\n"), err->message );
	    return FALSE; /* this will never happen because g_error exits */
	} else if ( *argc != 1 ) {
	    g_error( _("unexpected argument: %s\n"), (*argv)[1] );
	    return FALSE; /* this will never happen because g_error exits */
	}
	if ( !bdebug ) {
	    g_log_set_handler( G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG,
			       glf_dummy, NULL );
	}

	g_option_context_free( context );
    }

    *gui_context = gc;

    return TRUE;
}

/**
 * Do the main gui loop.
 *
 * @param gui_context pointer to a gui gui_context.
 * @return TRUE if ok, FALSE on error
 */
gboolean gui_do(GuiContext *gui_context) {
    GError *err = NULL;
    gint result;

    g_assert( gui_context );

    if ( !gui_context->inputvideo ) {
	g_error( _("No input video selected!\n") );
	return FALSE;
    }
    if ( !gui_context->outputdir ) {
	gchar *s_h, *s_hh;
	gui_context->outputdir = g_strdup( gui_context->inputvideo );
	for ( s_h = s_hh = gui_context->outputdir;
	      *s_hh;
	      s_hh++ )
	    if ( G_IS_DIR_SEPARATOR( *s_hh ) )
		s_h = s_hh;
	*s_h = 0;
    }
    if ( !gui_context->tmpdir ) {
	gui_context->tmpdir = g_strdup( g_get_tmp_dir() );
    }

    if ( ( result = ifile_set_filename(gui_context->inputvideo,
				       gui_context->iF_context, &err) ) < 0 ) {
	/* Something went wrong */
    }
    if ( err ) {
	gui_error_message( gui_context, &err );
	return FALSE; /* This will never happen because gui_error_message exits! */
    }

    if ( gui_context->title ) {
	ifile_set_title(gui_context->title,
			gui_context->iF_context,
			&err);
	if ( err ) {
	    g_printf( _("WARNING: Cannot set video title: %s\n"), err->message );
	    g_clear_error( &err );
	}
    }

    gui_context->progress = 0.0;
    if ( !createrec_do( gui_context->outputdir, 
			gui_context->number, gui_context->searchfree,
			gui_context->iF_context,
			gui_context->tmpdir,
			gui_context, &err ) ) {
	/* Something went wrong */
    }
    if ( err ) {
	gui_error_message( gui_context, &err );
	return FALSE; /* This will never happen because gui_error_message exits! */
    }

    return TRUE;
}

/**
 * Exit the gui
 *
 * @return TRUE if exit, FALSE if not
 */
gboolean gui_exit(GuiContext *gui_context) {
    if ( gui_context ) {
	if ( gui_context->iF_context )
	    ifile_exit( gui_context->iF_context );
	if ( gui_context->inputvideo )
	    g_free( gui_context->inputvideo );
	if ( gui_context->tmpdir )
	    g_free( gui_context->tmpdir );
	if( gui_context->outputdir )
	    g_free( gui_context->outputdir );
	if( gui_context->title )
	    g_free( gui_context->title );
	g_free( gui_context );
    }
    return TRUE;
}

/**
 * Spawn a process, show the output of the process and wait for termination
 *
 * Note: errors are not shown in the given gui context but reported via err!
 *
 * @param command The command to be executed
 * @param gui_context pointer to a gui context.
 * @param err the error context to be used
 * @return TRUE if everything was fine, FALSE on error
 */
gboolean gui_spawn_sync(const gchar *command, 
			GuiContext *gui_context, 
			GError **err ) {
    gchar **argv;
    gint argc, retval;
    gboolean bsuccess;

    if ( !g_shell_parse_argv( command, &argc, &argv, err ) )
	return FALSE;
    bsuccess = g_spawn_sync( NULL,
			     argv,
			     NULL,
			     G_SPAWN_SEARCH_PATH
			     |G_SPAWN_CHILD_INHERITS_STDIN,
			     NULL, /* spawn func */
			     NULL, /* user_data */
			     NULL, /* stdout (inherit) */
			     NULL, /* stderr (inherit) */
			     &retval,
			     err );
    g_free( argv );

    if ( !bsuccess )
	return FALSE;

    return ( WIFEXITED(retval) && !WEXITSTATUS(retval) );
}
