/*
 * gui-gtk.c - the gui for createrec-gtk
 * Copyright (c) 2009 Markus Kohm
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "gui.h"
#include "ifile.h"
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <sys/wait.h>

extern const gchar *gui_glade;

struct _gui_context {
    GtkWidget     *gW_main_dialog;
    GtkWidget     *gW_action_button;
    GtkWidget     *gW_inputfile_button;
    GtkFileFilter *ff_videopreferred,*ff_videoall;
    GtkWidget     *gW_tempdir_button;
    GtkWidget     *gW_outputdir_button;
    GtkWidget     *gW_dirnumber_button;
    GtkWidget     *gW_firstunuseddir_button;
    GtkWidget     *gW_quit_button;
    GtkWidget     *gW_status_progressbar;
    GtkWidget     *gW_status_label;
    GtkWidget     *gW_text_title;

    GtkWidget     *gW_about_dialog;

    IFileContext  *iF_context;
};

extern gboolean createrec_do( const gchar *dirname, 
			      guint number, gboolean searchfirstfree,
			      IFileContext *iF_context, const gchar *tmpdir,
			      GuiContext *gui_context, 
			      GError **err );

/**
 * Show a GError message
 */
static void gui_error_message( GuiContext *gui_context, GError **err ) {
    GtkWidget *gW_m;
    g_assert( gui_context );
    g_assert( err );
    g_assert( *err );

    gW_m = gtk_message_dialog_new( GTK_WINDOW( gui_context->gW_main_dialog ),
				   GTK_DIALOG_DESTROY_WITH_PARENT,
				   GTK_MESSAGE_ERROR,
				   GTK_BUTTONS_CANCEL,
				   (*err)->message );
    gtk_dialog_run( GTK_DIALOG( gW_m ) );
    gtk_widget_destroy( gW_m );
    
    g_clear_error( err );
}

/**
 * Show progress or state of the programm
 *
 * @param progress < 0 for not changing the progress, 
 *                 0.0..1.0 for changing progress.
 * @param message is a state string or NULL if state not changed
 * @param gui_context pointer to a gui context.
 */
void gui_show_state( gdouble progress, gchar *message,
		     GuiContext *gui_context ) {
    g_assert( gui_context != NULL
	      && gui_context->gW_status_progressbar != NULL
	      && gui_context->gW_status_label != NULL );
    if ( !(progress < 0.0) && !(progress > 1.0) ) {
	gchar *percent = g_strdup_printf( "%5.2f%%", 100*progress );
	gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR( gui_context->gW_status_progressbar ),
				       progress );
	gtk_progress_bar_set_text( GTK_PROGRESS_BAR( gui_context->gW_status_progressbar ),
				   percent );
	g_free( percent );
    }
    if ( message )
	gtk_label_set_text( GTK_LABEL( gui_context->gW_status_label ),
			    message );

    while (gtk_events_pending ())
	gtk_main_iteration ();
}

/**
 */
static void gui_set_sensitive( gboolean bsensitive, GuiContext *gui_context ) {
    GError *err;

    g_assert( gui_context
	      && gui_context->gW_action_button
	      && gui_context->gW_inputfile_button
	      && gui_context->gW_tempdir_button
	      && gui_context->gW_outputdir_button
	      && gui_context->gW_dirnumber_button
	      && gui_context->gW_firstunuseddir_button
	      && gui_context->gW_quit_button );

    if ( !bsensitive ) {

	gtk_widget_set_sensitive( gui_context->gW_action_button, FALSE );
	gtk_widget_set_sensitive( gui_context->gW_inputfile_button, FALSE );
	gtk_widget_set_sensitive( gui_context->gW_tempdir_button, FALSE );
	gtk_widget_set_sensitive( gui_context->gW_outputdir_button, FALSE );
	gtk_widget_set_sensitive( gui_context->gW_dirnumber_button, FALSE );
	gtk_widget_set_sensitive( gui_context->gW_firstunuseddir_button, FALSE );
	gtk_widget_set_sensitive( gui_context->gW_quit_button, FALSE );

    } else {
	gint nresult;
	const gchar *filename = gtk_file_chooser_get_filename( GTK_FILE_CHOOSER( gui_context->gW_inputfile_button ) );

	gtk_widget_set_sensitive( gui_context->gW_quit_button, TRUE );
	gtk_widget_set_sensitive( gui_context->gW_inputfile_button, TRUE );
	gtk_widget_set_sensitive( gui_context->gW_outputdir_button, TRUE );
	gtk_widget_set_sensitive( gui_context->gW_dirnumber_button, TRUE );
	gtk_widget_set_sensitive( gui_context->gW_firstunuseddir_button, TRUE );
	
	nresult = ifile_set_filename( filename, gui_context->iF_context, &err );
	if ( nresult < 0 ) {
	    g_clear_error( &err ); /* ignore the error message */
	    
	    gtk_file_chooser_set_filename( GTK_FILE_CHOOSER( gui_context->gW_inputfile_button ),
					   NULL ); /* clear filename */
	    gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_action_button),
				      FALSE ); 
	    gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_tempdir_button),
				      FALSE );
	} else if ( nresult > 0 ) {
	    /* No error but format conversation needed */
	    gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_action_button),
				      TRUE ); 
	    gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_tempdir_button),
				      TRUE ); 
	} else {
	    /* No error and no format conversation needed */
	    gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_action_button),
				      TRUE ); 
	    gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_tempdir_button),
				      FALSE ); 
	}
    }
}

/**
 * Signal handlers
 *
 * Note: All these are private (so they are not shown in gui.h), but must be
 *       exported for GtkBuilder usage.
 */

/**
 * If the about button in the main dialog was clicked, show the about dialog
 */
G_MODULE_EXPORT void about_button_clicked_cb( GtkButton *button,
					      GuiContext *gui_context) {
    g_assert( gui_context != NULL && gui_context->gW_about_dialog != NULL );
    gtk_dialog_run( GTK_DIALOG(gui_context->gW_about_dialog) );
    gtk_widget_hide( GTK_WIDGET(gui_context->gW_about_dialog) );
}

/**
 * If the action button in the main dialog was clicked, do the whole i/o job
 */
G_MODULE_EXPORT void action_button_clicked_cb( GtkButton *button,
					       GuiContext *gui_context) {
    GError *err = NULL;
    gchar *dirname;
    const gchar *tmpdir, *title;
    guint number;
    gboolean searchfree;

    g_assert( gui_context != NULL 
	      && gui_context->gW_outputdir_button != NULL
	      && gui_context->gW_inputfile_button != NULL 
	      && gui_context->gW_firstunuseddir_button != NULL );

    gui_show_state( 0.0, _("Start processing:"), gui_context );
    gui_set_sensitive( FALSE, gui_context );

    dirname = gtk_file_chooser_get_filename( GTK_FILE_CHOOSER( gui_context->gW_outputdir_button ) );
    if ( !dirname ) {
	/* Try to use the folder of the input file for the output */
	dirname = gtk_file_chooser_get_current_folder( GTK_FILE_CHOOSER( gui_context->gW_inputfile_button ) );
	if ( !dirname ) {
	    GtkWidget *gW_m;
	    gint nresponse;

	    gW_m = gtk_message_dialog_new( GTK_WINDOW( gui_context->gW_main_dialog ),
					   GTK_DIALOG_DESTROY_WITH_PARENT,
					   GTK_MESSAGE_ERROR,
					   GTK_BUTTONS_OK|GTK_BUTTONS_CANCEL,
					   _("Can't recognize valid output directory!\nYou should select an output directory now.") );
	    nresponse = gtk_dialog_run( GTK_DIALOG( gW_m ) );
	    gtk_widget_destroy( gW_m );
	    gui_set_sensitive( TRUE, gui_context );

	    if ( nresponse == GTK_RESPONSE_OK )
		gtk_widget_activate( gui_context->gW_outputdir_button );
	    return;
	}
    }

    ifile_set_title( gtk_entry_get_text( GTK_ENTRY( gui_context->gW_text_title ) ),
		     gui_context->iF_context, &err );

    tmpdir = gtk_file_chooser_get_filename( GTK_FILE_CHOOSER( gui_context->gW_tempdir_button ) );
    if ( !tmpdir )
	tmpdir = g_get_tmp_dir();

    number = (guint)gtk_spin_button_get_value( GTK_SPIN_BUTTON( gui_context->gW_dirnumber_button ) );
    if ( number < 0 || number > 9999 )
	g_error( _("directory number must be 0..9999!\n") );
    searchfree = gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( gui_context->gW_firstunuseddir_button ) );

    if ( !createrec_do( dirname, number, searchfree,
			gui_context->iF_context, tmpdir,
			gui_context, &err ) ) {
	/* Something went wrong! */
    }
    if ( err ) {
	gui_error_message( gui_context, &err );
    }	

    g_free( dirname );
    gui_set_sensitive( TRUE, gui_context );
}

/**
 * If the quit button in the main dialog was clicked or the main dialog was
 * quit by another action, exit from gui (and mostly from application)
 */
G_MODULE_EXPORT void quit_button_clicked_cb( GtkButton *button,
					     GuiContext *gui_context ) {
    gui_exit( gui_context );
}

/**
 * If the input file changed, check if the new input file may be used for
 * input and cancel the change if not.
 */
G_MODULE_EXPORT void inputfile_button_file_activated_cb( GtkWidget *button,
							 GuiContext *gui_context ) {
    gint nresult;
    GError *err = NULL;
    gchar *old_file;
    const gchar *new_file;
    g_assert( gui_context != NULL );

    new_file = ifile_get_filename( gui_context->iF_context );
    if ( new_file )
	old_file = g_strdup( new_file );
    else
	old_file = NULL;

    new_file = gtk_file_chooser_get_filename( GTK_FILE_CHOOSER( gui_context->gW_inputfile_button ) );
    if ( !new_file ) {
	if ( gui_context->gW_action_button )
	    gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_action_button),
				      FALSE );
	if ( old_file )
	    g_free( old_file );
	return;
    }

    nresult = ifile_set_filename( new_file, gui_context->iF_context, &err );
    if ( nresult < 0 ) {
	gui_error_message( gui_context, &err );

	if ( ! gtk_file_chooser_set_filename( GTK_FILE_CHOOSER( gui_context->gW_inputfile_button ),
					      old_file ) ) {
	    /* Cannot set old filename again! */
	    gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_action_button),
				      FALSE ); 
	    gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_tempdir_button),
				      FALSE );
	}
    } else if ( nresult > 0 ) {
	/* No error but format conversation needed */
	gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_action_button),
				  TRUE ); 
	gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_tempdir_button),
				  TRUE ); 
    } else {
	/* No error and no format conversation needed */
	gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_action_button),
				  TRUE ); 
	gtk_widget_set_sensitive( GTK_WIDGET(gui_context->gW_tempdir_button),
				  FALSE ); 
    }

    if ( nresult >= 0 ) {
	const gchar *title = ifile_get_title( gui_context->iF_context, &err );
	gtk_entry_set_text( GTK_ENTRY(gui_context->gW_text_title), title );
    }

    if ( old_file )
	g_free( old_file );
}

/**
 * Do all initialization for the gui without entering the main loop.
 *
 * Note, that argc and argv are input and output parameters.  After
 * the initialization all gui parameters are removed from the array.
 * You should call this only once!
 * 
 * @param argc pointer to number of arguments
 * @param argv array of arguments
 * @param gui_context pointer to a gui gui_context.
 * @return TRUE if ok, FALSE on error
 */
gboolean gui_init(int *argc, char ***argv, GuiContext **gui_context) {
    static gboolean onceonly = FALSE;
    GtkBuilder *builder;
    GError *err = NULL;
    GuiContext *gc;
 
    g_assert( argc != NULL && argv != NULL && !onceonly );
    g_assert( gui_context != NULL );
    gc = g_malloc0( sizeof( GuiContext ) );
    onceonly = TRUE;

    gtk_init (argc, argv);

    /* Create builder and load interface */
    builder = gtk_builder_new();
    gtk_builder_set_translation_domain( builder, PACKAGE );
#if DONT_USE_GLADE_STRING
#else
    if ( !gtk_builder_add_from_string( builder, gui_glade, -1, &err ) )
#endif
    if ( !gtk_builder_add_from_file( builder, "createrec.glade", &err ) )
	g_error( "Kann createrec.glade nicht laden: %s\n", err->message );

    if ( !ifile_init(argc, argv, &(gc->iF_context), &err) )
	g_error( "ToDO: add handler for ifile_init: %s\n", err->message );

    gc->ff_videoall = gtk_file_filter_new();
    gtk_file_filter_set_name( gc->ff_videoall, _("All video files") );
    gtk_file_filter_add_mime_type( gc->ff_videoall, "video/*" );
    gtk_file_filter_add_pattern( gc->ff_videoall, "*.ts" );
    gtk_file_filter_add_pattern( gc->ff_videoall, "*.mpg" );
    gtk_file_filter_add_pattern( gc->ff_videoall, "*.m2t" );
    gtk_file_filter_add_pattern( gc->ff_videoall, "*.m2p" );
    gc->ff_videopreferred = gtk_file_filter_new();
    gtk_file_filter_set_name( gc->ff_videopreferred, _("Prefered video files") );
    gtk_file_filter_add_mime_type( gc->ff_videopreferred, "video/mpegts" );
    gtk_file_filter_add_mime_type( gc->ff_videopreferred, "video/x-mpegts" );
    gtk_file_filter_add_pattern( gc->ff_videopreferred, "*.ts" );
    gtk_file_filter_add_pattern( gc->ff_videopreferred, "*.m2t" );

    gc->gW_main_dialog = GTK_WIDGET(gtk_builder_get_object(builder, 
							   "main_dialog"));
    gc->gW_action_button = GTK_WIDGET(gtk_builder_get_object(builder,
							     "action_button"));
    gc->gW_tempdir_button = GTK_WIDGET(gtk_builder_get_object(builder,
							      "tempdir_button"));
    gc->gW_inputfile_button = GTK_WIDGET(gtk_builder_get_object(builder,
								"inputfile_button"));
    gc->gW_outputdir_button = GTK_WIDGET(gtk_builder_get_object(builder,
								"outputdir_button"));
    gc->gW_dirnumber_button = GTK_WIDGET(gtk_builder_get_object(builder,
								"dirnumber_button"));
    gc->gW_firstunuseddir_button = GTK_WIDGET(gtk_builder_get_object(builder,
								     "firstunuseddir_button"));
    gc->gW_quit_button = GTK_WIDGET(gtk_builder_get_object(builder,
							   "quit_button"));
    gc->gW_status_progressbar = GTK_WIDGET(gtk_builder_get_object(builder,
								  "status_progressbar"));
    gc->gW_status_label = GTK_WIDGET(gtk_builder_get_object(builder,
							    "status_label"));
    gc->gW_text_title = GTK_WIDGET(gtk_builder_get_object(builder,
							  "text_title"));

    gtk_file_chooser_set_current_folder( GTK_FILE_CHOOSER( gc->gW_tempdir_button ),
					 g_get_tmp_dir() );

    gtk_file_chooser_add_filter( GTK_FILE_CHOOSER( gc->gW_inputfile_button ),
				 gc->ff_videopreferred );
    gtk_file_chooser_add_filter( GTK_FILE_CHOOSER( gc->gW_inputfile_button ),
				 gc->ff_videoall );

    gc->gW_about_dialog = GTK_WIDGET(gtk_builder_get_object(builder,
							    "about_dialog"));
    gtk_about_dialog_set_version( GTK_ABOUT_DIALOG( gc->gW_about_dialog ),
				  VERSION );
    
    gtk_builder_connect_signals( builder, gc );

    /* builder not longer needed */
    g_object_unref( builder );

    *gui_context = gc;

    return TRUE;
}

/**
 * Do the main gui loop.
 *
 * @param gui_context pointer to a gui gui_context.
 * @return TRUE if ok, FALSE on error
 */
gboolean gui_do(GuiContext *gui_context) {
    g_assert( gui_context != NULL && gui_context->gW_main_dialog != NULL );

    gtk_widget_show_all( gui_context->gW_main_dialog );
    gtk_main();

    return TRUE;
}

/**
 * Exit the gui
 *
 * @return TRUE if exit, FALSE if not
 */
gboolean gui_exit(GuiContext *gui_context) {
    gtk_main_quit();

    if ( gui_context ) {
	if ( gui_context->iF_context )
	    ifile_exit( gui_context->iF_context );
	g_free( gui_context );
    }
    return TRUE;
}

/**
 * Private data and methods needed for spawning
 */
typedef struct _SpawnData SpawnData;
struct _SpawnData {
    GtkDialog *dialog;
    GtkTextBuffer *out;
    GtkTextBuffer *err;
    GtkProgressBar *progress;
    gint timeout_id;
    gint status;
    GPid pid;
};

static void cb_child_watch( GPid      pid,
			    gint      status,
			    SpawnData *data ) {
    /* Remove timeout callback */
    g_source_remove( data->timeout_id );
    
    /* Close pid */
    g_spawn_close_pid( pid );

    /* Destroy dialog */
    gtk_widget_destroy( GTK_WIDGET(data->dialog) );

    /* save the status */
    data->status = status;
}

static gboolean cb_channel_watch( GIOChannel    *channel,
			          GIOCondition  cond,
			          GtkTextBuffer *buffer ) {
    gchar *string;
    gsize  size;
    gint lines;
    GtkTextIter iter_end;
 
    if( cond == G_IO_HUP ) {
        g_io_channel_unref( channel );
        return( FALSE );
    }

    g_io_channel_read_line( channel, &string, &size, NULL, NULL );
    if ( size > 80 ) {
	string[79] = '\n';
	string[80] = '\0';
	size = 80;
    }
    
    if ( ( lines = gtk_text_buffer_get_line_count( GTK_TEXT_BUFFER(buffer) ) )
	 > 10 ) {
	GtkTextIter iter_start;
	gtk_text_buffer_get_start_iter( GTK_TEXT_BUFFER(buffer),
					&iter_start );
	gtk_text_buffer_get_iter_at_line( GTK_TEXT_BUFFER(buffer),
					  &iter_end,
					  1 );
	gtk_text_buffer_delete( GTK_TEXT_BUFFER(buffer), 
				&iter_start,
				&iter_end );
    }

    gtk_text_buffer_get_end_iter( GTK_TEXT_BUFFER(buffer), &iter_end );
    gtk_text_buffer_insert( GTK_TEXT_BUFFER(buffer), &iter_end, string, size );
    
    g_free( string );

    while (gtk_events_pending ())
	gtk_main_iteration ();

    return( TRUE );
}

static gboolean cb_timeout( GtkProgressBar *progress )
{
    /* Bounce progress bar */
    gtk_progress_bar_pulse( GTK_PROGRESS_BAR(progress) );

    while (gtk_events_pending ())
	gtk_main_iteration ();

    return( TRUE );
}

/**
 * Spawn a process, show the output of the process and wait for termination
 *
 * Note: errors are not shown in the given gui context but reported via err!
 *
 * @param command The command to be executed
 * @param gui_context pointer to a gui context.
 * @param err the error context to be used
 * @return TRUE if everything was fine, FALSE on error
 */
gboolean gui_spawn_sync(const gchar *command, 
			GuiContext *gui_context, 
			GError **err ) {
    GPid pid;
    gchar **argv;
    gint argc, out_dev, err_dev, retval;
    GIOChannel *out_ch, *err_ch;
    SpawnData data;
    GtkWidget *content,
	*action,
	*text;
    gboolean bsuccess;

    if ( !g_shell_parse_argv( command, &argc, &argv, err ) )
	return FALSE;

    data.dialog = GTK_DIALOG(gtk_dialog_new());

#if 0
    content = gtk_dialog_get_content_area (GTK_DIALOG (data.dialog));
    action = gtk_dialog_get_action_area (GTK_DIALOG (data.dialog));
#else
    content=data.dialog->vbox;
    action=data.dialog->action_area;
#endif

    gtk_box_set_homogeneous( GTK_BOX( content ), FALSE );
    gtk_box_set_homogeneous( GTK_BOX( action ), FALSE );

    text = gtk_label_new( command );
    gtk_box_pack_start_defaults( GTK_BOX( content ), text );

    data.progress = GTK_PROGRESS_BAR(gtk_progress_bar_new());
    gtk_box_pack_start_defaults( GTK_BOX( content ), GTK_WIDGET( data.progress ) );
    
    text = gtk_text_view_new();
    data.out = gtk_text_view_get_buffer( GTK_TEXT_VIEW( text ) );
    gtk_box_pack_start( GTK_BOX( content ), text,
			 FALSE, FALSE, 4);

    text = gtk_hruler_new();
    gtk_box_pack_start( GTK_BOX( content ), text,
			FALSE, FALSE, 4);

    text = gtk_text_view_new();
    data.err = gtk_text_view_get_buffer( GTK_TEXT_VIEW( text ) );
    gtk_box_pack_start_defaults( GTK_BOX( content ), text );

    data.status = 0;
    
    bsuccess = g_spawn_async_with_pipes( NULL,
					 argv,
					 NULL,
					 G_SPAWN_DO_NOT_REAP_CHILD 
					 | G_SPAWN_SEARCH_PATH,
					 NULL,
					 NULL,
					 &data.pid,
					 NULL,
					 &out_dev,
					 &err_dev,
					 err );
    g_free( argv );

    if ( !bsuccess )
	return FALSE;

   /* Create channels that will be used to read data from pipes. */
#ifdef G_OS_WIN32
    out_ch = g_io_channel_win32_new_fd( out_dev );
    err_ch = g_io_channel_win32_new_fd( err_dev );
#else
    out_ch = g_io_channel_unix_new( out_dev );
    err_ch = g_io_channel_unix_new( err_dev );
#endif

    /* Add watches to channels */
    g_io_add_watch( out_ch, 
		    G_IO_IN | G_IO_HUP, 
		    (GIOFunc)cb_channel_watch, 
		    data.out );
    g_io_add_watch( err_ch, G_IO_IN | G_IO_HUP, 
		    (GIOFunc)cb_channel_watch, 
		    data.err );

    /* Install timeout fnction that will move the progress bar */
    data.timeout_id = g_timeout_add( 100, (GSourceFunc)cb_timeout, data.progress );

    g_child_watch_add( data.pid, (GChildWatchFunc)cb_child_watch, &data );

    gtk_widget_show_all( GTK_WIDGET(data.dialog) );

    if ( gtk_dialog_run(GTK_DIALOG(data.dialog)) != GTK_RESPONSE_NONE )
	gtk_widget_destroy(GTK_WIDGET(data.dialog));

    return ( WIFEXITED(data.status) && !WEXITSTATUS(data.status) );
}
