/*
 * ifile.c - the ifile for createrec-gtk
 * Copyright (c) 2009 Markus Kohm
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define _FILE_OFFSET_BITS 64

#include "ifile.h"
#include <glib/gi18n.h>
#include <libavformat/avformat.h>
#include <libavcodec/avcodec.h>

struct _ifile_context {
    const gchar     *filename;   /* Name of the input file or NULL */
    const gchar     *fileformat; /* Name of the input file format or NULL */
    const gchar     *title;      /* Title of the movie */
    AVFormatContext *pFormatCtx; /* valid if open, NULL if closed */
    guint16         PCR_PID;     /* PID of PCR packets */
    guint16         PMT_PID;     /* PID of packet with PMT */
    guint16         NID;         /* Network ID */
    guint8          stream_count;/* Number of streams used 0..8 */
    guint8          stream_type[8]; /* Types of streams */
    guint16         stream_pid[8]; /* PID of stream */
    guint32         packets;     /* number of packets after conversation */
    guint32         durationms;  /* duration in 1/1000 seconds */
};

/**
 * Open the file from the file context (if not already open)
 *
 * Note: You also have to close the file!
 *
 * @param iF_context the context of the file, that should be opened
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return TRUE if the file was opened, FALSE on error
 */
gboolean ifile_open(IFileContext *iF_context, GError **err) {
    gint retval;

    g_assert( iF_context );

    if ( !iF_context->pFormatCtx ) {
	av_register_all(); /* should not but may be called serveral times */

	if ( (retval = av_open_input_file(&iF_context->pFormatCtx, 
					  iF_context->filename, 
					  NULL, 
					  FFM_PACKET_SIZE, 
					  NULL )) < 0 ) {
	    g_set_error( err,
			 AV_STREAM_ERROR,
			 retval,
			 _("Can't open input file: %s: %s"),
			 iF_context->filename, 
			 g_strerror( AVUNERROR(retval) ) );
	    return FALSE;
	} else if ( !iF_context->pFormatCtx->nb_streams &&
		    (retval = av_find_stream_info(iF_context->pFormatCtx)) < 0 ) {
	    g_set_error( err,
			 AV_STREAM_ERROR,
			 retval,
			 _("Can't find stream infos: %s: %s"),
			 iF_context->filename,
			 g_strerror( AVUNERROR(retval) ) );
	    av_close_input_file(iF_context->pFormatCtx);
	    iF_context->pFormatCtx = NULL;
	    return FALSE;
	}
    }

    return TRUE;
}

/**
 * Close the file from the file context (if open)
 *
 * Note: The file context is still valid after closing. You have to use
 *       ifile_exit zu destroy it.
 *
 * @param iF_context the context of the file, that should be closed
 */
void ifile_close(IFileContext *iF_context) {
    g_assert( iF_context );

    if ( iF_context->pFormatCtx ) {
	av_close_input_file( iF_context->pFormatCtx );
	iF_context->pFormatCtx = NULL;
    }
}

/**
 * Get the current file name from input file context.
 *
 * @param iF_context the input file context.
 * @result either the file name from the given input file context or NULL
 */
const gchar *ifile_get_filename( IFileContext *iF_context ) {
    g_assert( iF_context != NULL );
    return iF_context->filename;
}

/**
 * Ask for the format of the file.
 *
 * Note: If the file was currently not opened, it will be opened temporary 
 *       only.
 *
 * @param iF_context the context of the file.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return name string of the file format or NULL on error.
 */
const gchar *ifile_get_format(IFileContext *iF_context, GError **err) {
    g_assert( iF_context );

    if ( !iF_context->fileformat ) {
	gboolean bclose = !iF_context->pFormatCtx;

	if ( !ifile_open(iF_context, err) )
	    return NULL;

	iF_context->fileformat = g_strdup(iF_context->pFormatCtx->iformat->name);
	if ( bclose )
	    ifile_close( iF_context );
    }

    return iF_context->fileformat;
}

/**
 * Exit the ifile
 *
 * Note, that iF_context is still valid after the call. Use ifile_exit() to
 * additionaly free iF_context.
 *
 * @param iF_context pointer to a ifile context.
 * @return TRUE if exit, FALSE if not
 */
gboolean ifile_exit0(IFileContext *iF_context) {
    if ( iF_context ) {
	ifile_close( iF_context );
	if ( iF_context->filename ) {
	    g_free( (gpointer)iF_context->filename );
	    iF_context->filename = NULL;
	}
	if ( iF_context->fileformat ) {
	    g_free( (gpointer)iF_context->fileformat );
	    iF_context->fileformat = NULL;
	}
	if ( iF_context->title ) {
	    g_free( (gpointer)iF_context->title );
	    iF_context->title = NULL;
	}
    }
    return TRUE;
}

/**
 * Set the current file name and do all needed initializations for input.
 *
 * @param filename the new filename to be used for input.
 * @param iF_context the input file context.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @result < 0 if the file name cannot be used for input, 
 *         = 0 if the file name can be used for direct input, 
 *         > 0 if the file name can be used for input after conversation 
 *             of the input format.
 */
gint ifile_set_filename( const gchar *filename, 
			 IFileContext *iF_context, GError **err ) {
    gint retval;
    IFileContext *iF_c;

    g_assert( iF_context && filename );

    ifile_close( iF_context );

    if ( !ifile_init( NULL, NULL, &iF_c, err ) )
	return -1;

    iF_c->filename = g_strdup( filename );

    if ( !ifile_get_format( iF_c, err ) ) {
	ifile_exit( iF_c );
	return -1;
    }

    /* We may use the file name for input: */
    ifile_exit0( iF_context );
    memcpy( iF_context, iF_c, sizeof( IFileContext ) );

    return abs(g_strcmp0( iF_context->fileformat, "mpegts" ));
}

/**
 * get the title of the file
 *
 * @param iF_context the input file context.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return title or NULL
 */
const gchar *ifile_get_title( IFileContext *iF_context, GError **err ) {
    const gchar *title = iF_context->title;

    if ( !ifile_open(iF_context, err) )
	return NULL;

    if ( !title ) {
	AVMetadataTag *t = av_metadata_get( iF_context->pFormatCtx->metadata,
					    "title", NULL, 0 );
	if ( t )
	    title = t->value;
    }
    if ( !title || !*title ) {
	title = iF_context->filename;
	if ( title ) {
	    const gchar *h;
	    for ( h = title; *h; h++ ) {
		if ( G_IS_DIR_SEPARATOR( *h ) )
		    title = h+1;
	    }
	}
	if ( !title || !*title ) {
	    title = _("createrec movie");
	}
    }

    return title;
}

/**
 * set the title of the file
 *
 * @param title the new title of the file
 * @param iF_context the input file context.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return TRUE on success, FALSE on error
 */
gboolean ifile_set_title( const gchar *title, 
			  IFileContext *iF_context, GError **err ) {
    g_assert( iF_context );
    if ( iF_context->title )
	g_free( (gpointer)iF_context->title );
    iF_context->title = title ? g_strdup( title ) : NULL;
    return TRUE;
}

/**
 * Do all initialization for unknown input files.
 *
 * Note, that argc and argv are input and output parameters.  After
 * the initialization all gui parameters are removed from the array.
 *
 * @param argc pointer to number of arguments.
 * @param argv array of arguments.
 * @param iF_context pointer to a ifile context.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return TRUE if ok, FALSE on error.
 */
gboolean ifile_init(int *argc, char ***argv, 
		    IFileContext **iF_context, GError **err) {
    g_assert( iF_context != NULL );

    *iF_context = g_malloc0( sizeof( IFileContext ) );
    return TRUE;
}

/**
 * Ask for the number of streams of a given type at the file.
 *
 * Note: if the file was currently not opened, it will be.
 *
 * @param iF_context the context of the file.
 * @param codec_type the type of codec searching for.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return number of streams of the type at the file or -1 on error.
 */
static gint ifile_get_typed_stream_count(IFileContext *iF_context, 
					 enum CodecType codec_type,
					 GError **err) {
    guint n, cnt;

    g_assert( iF_context );

    if ( !ifile_open( iF_context, err ) )
	return -1;

    for ( n = cnt = 0;
	  n < iF_context->pFormatCtx->nb_streams;
	  n++ ) {
	if ( iF_context->pFormatCtx->streams[n]->codec->codec_type==codec_type )
	    cnt++;
    }

    return cnt;
}
gint ifile_get_video_stream_count(IFileContext *iF_context, GError **err) {
    return ifile_get_typed_stream_count(iF_context, CODEC_TYPE_VIDEO, err );
}
gint ifile_get_audio_stream_count(IFileContext *iF_context, GError **err) {
    return ifile_get_typed_stream_count(iF_context, CODEC_TYPE_AUDIO, err );
}
gint ifile_get_subtitle_stream_count(IFileContext *iF_context, GError **err) {
    return ifile_get_typed_stream_count(iF_context, CODEC_TYPE_SUBTITLE, err );
}

/**
 * Ask for the codec of a given stream type.
 *
 * Note: if the file was currently not opened, it will be.
 *
 * @param iF_context the context of the file.
 * @param codec_type the type of codec searching for.
 * @param stream_number The number of the stream with given type.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return name string of the codec or NULL on error or if not enough
 *              streams where found.
 */
static const gchar *ifile_get_typed_codec_name(IFileContext *iF_context, 
					       enum CodecType codec_type,
					       guint stream_number,
					       GError **err) {
    guint n, cnt;
    
    g_assert( iF_context );

    if ( !ifile_open( iF_context, err ) )
	return NULL;

    for ( n = cnt = 0;
	  n < iF_context->pFormatCtx->nb_streams;
	  n++ ) {
	if ( iF_context->pFormatCtx->streams[n]->codec->codec_type==codec_type ) {
	    if ( stream_number == cnt++ ) {
		return iF_context->pFormatCtx->streams[n]->codec->codec_name;
	    }
	}
    }

    return NULL;
}
const gchar *ifile_get_video_codec_name(IFileContext *iF_context, 
					guint stream_number,
					GError **err) {
    return ifile_get_typed_codec_name(iF_context,CODEC_TYPE_VIDEO,stream_number,err);
}
const gchar *ifile_get_audio_codec_name(IFileContext *iF_context, 
					guint stream_number,
					GError **err) {
    return ifile_get_typed_codec_name(iF_context,CODEC_TYPE_AUDIO,stream_number,err);
}

/**
 * Ask for the codec of a given stream type.
 *
 * Note: if the file was currently not opened, it will be.
 *
 * @param iF_context the context of the file.
 * @param codec_type the type of codec searching for.
 * @param stream_number The number of the stream with given type.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return name string of the codec or NULL on error or if not enough
 *              streams where found.
 */
enum CodecID ifile_get_typed_codec_id(IFileContext *iF_context, 
				      enum CodecType codec_type,
				      guint stream_number,
				      GError **err) {
    guint n, cnt;
    
    g_assert( iF_context );

    if ( !ifile_open( iF_context, err ) )
	return CODEC_ID_NONE;

    for ( n = cnt = 0;
	  n < iF_context->pFormatCtx->nb_streams;
	  n++ ) {
	if ( iF_context->pFormatCtx->streams[n]->codec->codec_type==codec_type ) {
	    if ( stream_number == cnt++ ) {
		return iF_context->pFormatCtx->streams[n]->codec->codec_id;
	    }
	}
    }

    return CODEC_ID_NONE;
}
enum CodecID ifile_get_video_codec_id(IFileContext *iF_context, 
				      guint stream_number,
				      GError **err) {
    return ifile_get_typed_codec_id(iF_context,CODEC_TYPE_VIDEO,stream_number,err);
}
enum CodecID ifile_get_audio_codec_id(IFileContext *iF_context, 
				      guint stream_number,
				      GError **err) {
    return ifile_get_typed_codec_id(iF_context,CODEC_TYPE_AUDIO,stream_number,err);
}


/**
 * Ask for the language of an stream with given type.
 *
 * Note: if the file was currently not opened, it will be.
 *
 * @param iF_context the context of the file.
 * @param codec_type the type of codec searching for.
 * @param stream_number The number of the stream with given type.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return name string of the language or NULL on error or if not enough
 *              streams where found.
 */
static const gchar *ifile_get_typed_lang(IFileContext *iF_context, 
					 enum CodecType codec_type,
					 guint stream_number,
					 GError **err) {
    guint n, cnt;
    
    g_assert( iF_context );

    if ( !ifile_open( iF_context, err ) )
	return NULL;

    for ( n = cnt = 0;
	  n < iF_context->pFormatCtx->nb_streams;
	  n++ ) {
	if ( iF_context->pFormatCtx->streams[n]->codec->codec_type==codec_type ) {
	    if ( stream_number == cnt++ ) {
		AVMetadataTag *lang = av_metadata_get( iF_context->pFormatCtx->streams[n]->metadata,
						       "language", NULL, 0 );
		return lang ? lang->value : NULL;
	    }
	}
    }

    return NULL;
}
const gchar *ifile_get_audio_lang(IFileContext *iF_context, 
				  guint stream_number,
				  GError **err) {
    return ifile_get_typed_lang(iF_context,CODEC_TYPE_AUDIO,stream_number,err);
}
const gchar *ifile_get_subtitle_lang(IFileContext *iF_context, 
				     guint stream_number,
				     GError **err) {
    return ifile_get_typed_lang(iF_context,CODEC_TYPE_SUBTITLE,stream_number,err);
}

/**
 * Exit the ifile
 *
 * Note, that iF_context is not longer valid after this call!
 *
 * @param iF_context pointer to a ifile context.
 * @return TRUE if exit, FALSE if not
 */
gboolean ifile_exit(IFileContext *iF_context) {
    if ( iF_context ) {
	ifile_exit0( iF_context );
	g_free( iF_context );
    }

    return TRUE;
}

/**
 * Note: rec should be another class, but to increase speed, we're using
 *       direct access to IFileContext. So IFileContext must be a friend.
 */
#include "rec.h"
#include <glib/gstdio.h>

#define MAX_SEARCH_PAT_PMT 10000

static gboolean analyze_pmt( const guint8 *packet,
			     IFileContext *iF_context,
			     GError **err ) {
    const guint8 *p, *section_start;
    guint8 table_id, pointer_field;
    guint8 adaptation_field_control;
    guint16 section_length;
    guint16 program_number, program_info_length;

    g_assert( packet );
    g_assert( iF_context );

    p = packet;
    p += 3; /* Sync + PID */
    adaptation_field_control = (getguint8( p++ ) >> 4) & 3;

    /* Jump over adaptation field */
    if ( adaptation_field_control & 2 ) {
	p += getguint8( p++ );
	p++;
    }

    /* Jump over pointer field */
    pointer_field = getguint8( p++ );
    p += pointer_field;

    table_id = getguint8( p++ );
    if ( table_id != 0x02 ) {
	g_set_error( err,
		     CR_STREAM_ERROR,
		     CR_STREAM_ERROR_TIDPID,
		     _("Packet with PMT-PID 0x0002 has table_id %u instead of 0x02"),
		     (guint)table_id );
	return FALSE;
    }

    section_length = getguint16( p );
    p += 2;
    section_start = p;
    if ( ( section_length & 0xc000 ) != 0x8000 ) {
	/* Note: payload_start_indicator only needed, because we cannot 
	         handle splitted tables currently */
	g_set_error( err,
		     CR_STREAM_ERROR,
		     CR_STREAM_ERROR_TABNVLD,
		     _("PMT invalid") );
	return FALSE;
    }
    section_length &= 0xfff;
    if (( section_length < 13 ) 
	|| ( section_length > ( 176 - pointer_field ))) {
	/* Note: The real max. value would be 1021 but we cannot handle
	         splitted tables currently! */
	g_set_error( err,
		     CR_STREAM_ERROR,
		     CR_STREAM_ERROR_TABNVLD,
		     _("PMT invalid") );
	return FALSE;
    }

    program_number = getguint16( p );
    p += 2;

    /* Note: currently version_number, current_next_indicator, 
             section_number and last_section_numbered are ignored */
    p += 3;

    iF_context->PCR_PID = getguint16( p ) & 0x1fff;
    p += 2;

    program_info_length = getguint16( p ) & 0x0fff;
    p += program_info_length + 2;

    iF_context->stream_count = 0;
    while ( p < section_start + section_length - 4 ) {
	guint8 stream_type;
	guint16 PID;
	guint16 ES_info_length;
	stream_type = getguint8( p++ );
	PID = getguint16( p ) & 0x1fff;
	p += 2;
	ES_info_length = getguint16( p ) & 0xfff;
	p += 2 + ES_info_length;
	
	if (( stream_type >= 1 ) && ( stream_type <= 4 )) { /* VIDEO/AUDIO */
	    if ( iF_context->stream_count < 8 ) { /* We can handle max. 8 streams */
		int n;
		for ( n = 0; 
		      (n < iF_context->stream_count) 
			  && (iF_context->stream_pid[n] != PID); 
		      n++ ) ;
		if ( n == iF_context->stream_count ) {
		    g_debug((stream_type < 3)
			    ? _("Stream %d with PID 0x%04x is VIDEO\n")
			    : _("Stream %d with PID 0x%04x is AUDIO\n"),
			    n, 
			    PID );
		    iF_context->stream_pid[n] = PID;
		    iF_context->stream_type[n] = stream_type;
		    iF_context->stream_count++;
		} else if ( iF_context->stream_type[n] != stream_type ) {
		    g_error( "FixMe: Cannot handle stream type change!\n" );
		}
	    }
	}
    }

    return TRUE;
}

static gboolean analyze_pat( const guint8 *packet,
			     IFileContext *iF_context,
			     GError **err ) {
    const guint8 *p, *section_start;
    guint8 table_id, pointer_field;
    guint8 adaptation_field_control;
    guint16 section_length;
    guint16 program_number, program_info_length;

    g_assert( packet );
    g_assert( iF_context );

    p = packet;
    p += 3; /* Sync + PID */
    adaptation_field_control = (getguint8( p++ ) >> 4) & 3;

    /* Jump over adaptation field */
    if ( adaptation_field_control & 2 ) {
	p += getguint8( p++ );
	p++;
    }

    /* Jump over pointer field */
    pointer_field = getguint8( p++ );
    p += pointer_field;

    table_id = getguint8( p++ );
    if ( table_id != 0x00 ) {
	g_set_error( err,
		     CR_STREAM_ERROR,
		     CR_STREAM_ERROR_TIDPID,
		     _("Packet with PAT-PID 0x0000 has table_id %u instead of 0x00"),
		     (guint)table_id );
	return FALSE;
    }

    section_length = getguint16( p );
    p += 2;
    section_start = p;
    if ( ( section_length & 0xc000 ) != 0x8000 ) {
	/* Note: payload_start_indicator only needed, because we cannot 
	         handle splitted tables currently */
	g_set_error( err,
		     CR_STREAM_ERROR,
		     CR_STREAM_ERROR_TABNVLD,
		     _("PAT invalid") );
	return FALSE;
    }
    section_length &= 0xfff;
    if (( section_length < 9 ) 
	|| ( section_length > ( 176 - pointer_field ))) {
	/* Note: The real max. value would be 1021 but we cannot handle
	         splitted tables currently! */
	g_set_error( err,
		     CR_STREAM_ERROR,
		     CR_STREAM_ERROR_TABNVLD,
		     _("PAT invalid") );
	return FALSE;
    }

    p += 2; /* Note: currently we ignore the transport stream id */
    p += 3; /* Note: currently version_number, current_next_indicator, 
	       section_number and last_section_number are ignored */

    while ( ( p < section_start + section_length - 4 ) 
	    && ( p < packet + 184 ) ) {
	guint16 program_number, PID;
	program_number = getguint16( p );
	p += 2;
	PID = getguint16( p ) & 0x1fff;
	p += 2;
	if ( program_number == 0 ) {
	    /* Network PID ignored! */
	    iF_context->NID = PID;
	} else {
	    iF_context->PMT_PID = PID;
	    return TRUE; /* ignore everything else */
	}
    }

    return FALSE;
}

static void init_null_packet(guint8 *packet, IFileContext *iF_context) {
    gint n, a_streams;
    const gchar *title = ifile_get_title( iF_context, NULL );
    memset( packet, 0, 188 );
    *packet++ = 0x47; /* Sync */
    *packet++ = 0x1f; /* PID */
    *packet++ = 0xff;
    *packet++ = 0x10; /* flags */
    *packet++ = 0x02; /* magic */
    *packet++ = 0x32;
    *packet++ = 0x23;
    *packet++ = 0x10;
    *packet++ = 0x01;
    packet += 7; /* reserved */
    packet += 4; /* info_packets */
    packet += 4; /* duration */
    *packet++ = iF_context->stream_count;
    packet += 3;
    for ( n = a_streams = 0; 
	  n < iF_context->stream_count; 
	  n++ ) {
	packet[n*2]   = iF_context->stream_pid[n] & 0xff;
	packet[n*2+1] = (iF_context->stream_pid[n] >> 8) & 0xff;
	packet[n+16]  = ( iF_context->stream_type[n] >= 0
			  && iF_context->stream_type[n] <= 2 ) ? 0x01 :
	    ( iF_context->stream_type[n] >= 3 
	      && iF_context->stream_type[n] <= 4 ) ? 0x02 : 0x00;
	if ( packet[n+16] == 0x02 ) {
	    const gchar *lang = ifile_get_audio_lang( iF_context, a_streams, NULL );
	    a_streams++;
	    if ( lang && *lang ) {
		if ( lang[1] ) {
		    if ( lang[2] ) {
			packet[n+24] = lang[2];
			packet[n+25] = lang[1];
			packet[n+26] = lang[0];
		    } else {
			packet[n+24] = lang[1];
			packet[n+25] = lang[0];
		    }
		} else {
		    packet[n+24] = lang[0];
		}
	    }
	}
	packet[n+27] = 0;
    }
    packet += 56;
    packet += 4;
    packet += 4;
    packet += 4;
    *packet++ = 0x02;
	
    strncpy( packet, title, 62 );
    packet += 63;
    packet += 12;
    *packet++ = 0x26;
    *packet++ = 0x57;
    packet += 2;
}

static void generate_null_packets(IFileContext *iF_context,
				  guint64 packet_count, guint32 durationms,
				  guint8 **header,
				  guint8 **info,
				  guint8 **empty) {
    guint8 *packet;
    if ( header ) {
	guint32 info_count = (guint32)((packet_count + 511) / 512);
	guint32 file_count = (info_count + 22309) / 22310;
	if ( !*header ) {
	    *header = g_malloc( 188 );
	    init_null_packet( *header, iF_context );
	}
	packet = *header;
	packet[0x10] = ( info_count ) & 0xff ;
	packet[0x11] = ( info_count >> 8 ) & 0xff;
	packet[0x12] = ( info_count >> 16 ) & 0xff;
	packet[0x13] = ( info_count >> 24 ) & 0xff;
	packet[0x14] = ( durationms ) & 0xff ;
	packet[0x15] = ( durationms >> 8 ) & 0xff;
	packet[0x16] = ( durationms >> 16 ) & 0xff;
	packet[0x17] = ( durationms >> 24 ) & 0xff;
	packet[0xb0] = ( file_count ) & 0xff ;
	packet[0xb1] = ( file_count >> 8 ) & 0xff;
	packet[0xb2] = ( file_count >> 16 ) & 0xff;
	packet[0xb3] = ( file_count >> 24 ) & 0xff;
    }
    if ( info ) {
	if ( !*info ) {
	    *info = g_malloc( 188 );
	    init_null_packet( *info, iF_context );
	}
	packet = *info;
	packet[0x54] = ( durationms ) & 0xff ;
	packet[0x55] = ( durationms >> 8 ) & 0xff;
	packet[0x56] = ( durationms >> 16 ) & 0xff;
	packet[0x57] = ( durationms >> 24 ) & 0xff;
    }
    if ( empty ) {
	if ( !*empty ) {
	    packet = g_malloc( 188 );
	    memset( packet, 0xff, 188 );
	    packet[0] = 0x47;
	    packet[1] = 0x1f;
	    packet[3] = 0x10;
	    *empty = packet;
	}
    }
}

/**
 * create new ts files from ifile
 *
 * Note: creating the ts will change the iF_context!
 *
 * @param dstdir The folder where to create the ts files.
 * @param iF_context pointer to the ifile context of the source
 * @param gui_context a gui context to report
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return TRUE on success, FALSE on error
 */
gboolean rec_create_ts(const gchar *dstdir,
		       IFileContext *iF_context,
		       GuiContext *gui_context,
		       GError **err) {
    FILE *in, *out = NULL;
    gchar *dstpath, *out_name;
    off_t in_size, o_read;
    guint64 n_packet, o_packet;
    guint8 o_file;
    int m, n;
    gboolean bresult = TRUE;
    gboolean bshow;
    guint64 duration[10];
    AVRational time_base[10];
    AVPacket avpacket;
    guint32 durationms;
    guint8 *packet = NULL, *pat_packet = NULL, *pmt_packet = NULL, 
	*null_h_packet = NULL, *null_i_packet = NULL, *null_e_packet = NULL;

    g_assert( iF_context && iF_context->filename );

    gui_show_state( 0.0, _("Creating rec.ts ..."), gui_context );

    dstpath = g_strndup( dstdir, strlen(dstdir) + 10 ); /* 8 would be enough */
    
    out_name = dstpath + strlen( dstpath );
    if ( out_name != dstpath
	 && !G_IS_DIR_SEPARATOR( out_name[-1] ) )
	*out_name++ = G_DIR_SEPARATOR;
    strcpy( out_name, "rec.ts" ); /* That's always the first file name! */

    if ( ( in = g_fopen( iF_context->filename, "rb" ) ) == NULL 
	 || fseeko( in, 0, SEEK_END )
	 || ( in_size = ftello( in ) ) == -1
	 || fseeko( in, 0, SEEK_SET ) ) {
	if ( in )
	    fclose( in );
	g_free( dstpath );
	g_set_error( err,
		     G_FILE_ERROR,
		     g_file_error_from_errno( errno ),
		     _("Can't read video file: %s: %s"),
		     iF_context->filename,
		     g_strerror( errno ) );
	return FALSE;
    }

    if ( !ifile_open( iF_context, err ) ) {
	fclose( in );
	g_free( dstpath );
	return FALSE;
    }

    gui_show_state( 0.0, _("Searching PAT and PMT ..."), gui_context );
    packet = g_malloc( 188 );
    n_packet = 0;
    if (!iF_context->PMT_PID)
	iF_context->PMT_PID = 0x02;

    for ( n_packet = 0;
	  n_packet < MAX_SEARCH_PAT_PMT
	      && ( !pat_packet || !pmt_packet )
	      && ( fread( packet, 188, 1, in ) == 1 );
	  n_packet++ ) {
	guint16 pid;
	if ( packet[0] != 0x47 ) { /* packet not in sync */
	    guint8 *p;
	    gchar *message = g_strdup_printf( _("Packet %llu not in sync"), 
					      (unsigned long long)n_packet );
	    gui_show_state( -1.0, message, gui_context );
	    g_free( message );
	    while ( (p = memchr( packet, 0x47, 188 )) == NULL
		    && ( fread( packet, 188, 1, in ) == 1 ) );
	    if ( !p ) /* packet still not in sync */
		break;
	    g_memmove( packet, p, 188 - ( p - packet ) );
	    if ( fread( packet, p-packet, 1, in ) != 1 )
		break;
	}
	pid = getguint16( packet + 1 );
	if ( ( pid & 0x8000 ) /* error */
	     || !( pid & 0x4000 ) /* !payload_unit_start_indicator */
	    )
	    continue;
	pid &= 0x1fff;
	if ( !pmt_packet && pid == iF_context->PMT_PID ) { /* PMT */
	    pmt_packet = packet;
	    packet = g_malloc( 188 );
	    if ( !analyze_pmt( pmt_packet, iF_context, err ) ) {
		bresult = FALSE;
		goto cleanup;
	    }
	} else if ( !pat_packet && pid == 0x000 ) { /* PAT */
	    if ( !analyze_pat( packet, iF_context, err ) ) {
		if ( err ) {
		    bresult = FALSE;
		    goto cleanup;
		}
	    } else {
		pat_packet = packet;
		packet = g_malloc( 188 );
		n_packet = 0;              /* if first PMT before PAT */
		fseeko( in, 0, SEEK_SET ); /* if first PMT before PAT */
	    }
	}
    }

    fseeko( in, 0, SEEK_SET );

    if ( !pmt_packet || !pat_packet )
	g_error( "FixMe: Cannot handle streams without pmt or pat!\n" );

    for( n = 0; n < iF_context->pFormatCtx->nb_streams && n < 10; n++ ) {
	duration[n] = 0;
	time_base[n].num = 1;
	time_base[n].den = 1;
    }

    avpacket.stream_index = 10000;
    avpacket.duration = 0;

    /* Main packet loop */
    for ( n_packet = o_packet = 0, o_read = 0, o_file = 0, durationms=0;
	  fread( packet, 188, 1, in ) == 1;
	  n_packet++ ) {
	guint16 pid;
	gchar *comment;
	o_read += 188;
	if ( packet[0] != 0x47 ) { /* packet not in sync */
	    guint8 *p;
	    gchar *message = g_strdup_printf( _("Packet %llu not in sync"), 
					      (unsigned long long)n_packet );
	    gui_show_state( -1.0, message, gui_context );
	    g_free( message );
	    while ( (p = memchr( packet, 0x47, 188 )) == NULL
		    && ( fread( packet, 188, 1, in ) == 1 ) )
		o_read += 188;
	    if ( !p ) /* packet still not in sync */
		break;
	    g_memmove( packet, p, packet + 188 - p );
	    if ( fread( packet, p-packet, 1, in ) != 1 )
		break;
	    o_read += p-packet;
	}
	
	if ( o_read > in_size ) {
	    g_warning( _("File size increased while reading!\n") );
	    fseeko( in, 0, SEEK_END );
	    in_size = ftello( in );
	    fseeko( in, o_read, SEEK_SET );
	}

	pid = getguint16( packet + 1 ) & 0x1fff;

	for ( n = 0; 
	      n < iF_context->stream_count
		  && pid != iF_context->stream_pid[n];
	      n++ );

	if ( n < iF_context->stream_count ) {

	    /* calc duration */
	    while ( url_ftell(iF_context->pFormatCtx->pb) < o_read ) {
		if ( avpacket.stream_index < 10 ) {
		    duration[avpacket.stream_index] += avpacket.duration;
		    av_free_packet( &avpacket );
		    avpacket.stream_index = 10000;
		}
		if ( av_read_frame(iF_context->pFormatCtx,&avpacket) < 0 )
		    break;
	    }

	    for( n = 0; 
		 n < iF_context->pFormatCtx->nb_streams && n < 10; 
		 n++ ) {
		guint64 d = duration[n] * 1000 
		    * iF_context->pFormatCtx->streams[n]->time_base.num
		    / iF_context->pFormatCtx->streams[n]->time_base.den;
		if ( d > durationms )
		    durationms = d;
	    }

	    if ( ( o_packet % 512 ) == 0 ) {
		/* write every 512 packets a new comag info block */
		bshow = TRUE; /* show new state */

		if ( !( o_packet % (22310LL * 512) ) ) {
		    /* start every 22310 info blocks a new file */
		    if ( out ) {
			/* close current output file */
			if ( fflush( out ) ) {
			    g_set_error( err,
					 G_FILE_ERROR,
					 g_file_error_from_errno (errno),
					 _("File can't be written correctly: %s: %s"),
					 dstpath,
					 g_strerror( errno ) );
			    bresult = FALSE;
			    goto cleanup;
			} else
			    fclose( out );
			out = NULL;
			g_sprintf( out_name, "rec.%02u", ++o_file );
		    }
		    /* open new output file */
		    comment = g_strdup_printf( _("Writing %s ..."), out_name );
		    gui_show_state( -1.0, comment, gui_context );
		    g_free( comment );
		    if ( ( out = fopen( dstpath, "wb" ) ) == NULL ) {
			g_set_error( err,
				     G_FILE_ERROR,
				     g_file_error_from_errno (errno),
				     _("Can't open file: %s: %s"),
				     dstpath,
				     g_strerror( errno ) );
			bresult = FALSE;
			goto cleanup;
		    }
		}
		
		if ( o_packet == 0 ) {
		    /* first info block is a header */
		    generate_null_packets( iF_context, o_packet, 0,
					   &null_h_packet, &null_i_packet,
					   &null_e_packet );
		    if ( fwrite( null_h_packet, 188, 1, out ) != 1
			 || fwrite( null_e_packet, 188, 1, out ) != 1
			 || fwrite( pmt_packet, 188, 1, out ) != 1
			 || fwrite( pat_packet, 188, 1, out ) != 1
			 || fwrite( pmt_packet, 188, 1, out ) != 1 ) {
			g_set_error( err,
				     G_FILE_ERROR,
				     errno ? g_file_error_from_errno(errno)
				     : G_FILE_ERROR_IO,
				     _("File can't be written correctly: %s: %s"),
				     dstpath,
				     g_strerror( errno ? errno : EIO ) );
			bresult = FALSE;
			goto cleanup;
		    }
		    o_packet += 5;
		} else {
		    guint8 cb;

		    cb = pat_packet[3];
		    cb = ( cb & 0xf0 ) | ( ( cb + 1 ) & 0xf );
		    pat_packet[3] = cb;

		    cb = pmt_packet[3];
		    cb = ( cb & 0xf0 ) | ( ( cb + 1 ) & 0xf );
		    pmt_packet[3] = cb;

		    generate_null_packets( iF_context, o_packet, durationms,
					   NULL, &null_i_packet, NULL );
		    
		    if ( fwrite( null_i_packet, 188, 1, out ) != 1 
			 || fwrite( pat_packet, 188, 1, out ) != 1
			 || fwrite( pmt_packet, 188, 1, out ) != 1 ) {
			g_set_error( err,
				     G_FILE_ERROR,
				     errno ? g_file_error_from_errno(errno)
				     : G_FILE_ERROR_IO,
				     _("File can't be written correctly: %s: %s"),
				     dstpath,
				     g_strerror( errno ? errno : EIO ) );
			bresult = FALSE;
			goto cleanup;
		    }
		    o_packet += 3;
		}
	    }

	    if ( fwrite( packet, 188, 1, out ) != 1 ) {
		g_set_error( err,
			     G_FILE_ERROR,
			     errno ? g_file_error_from_errno(errno)
			     : G_FILE_ERROR_IO,
			     _("File can't be written correctly: %s: %s"),
			     dstpath,
			     g_strerror( errno ? errno : EIO ) );
		bresult = FALSE;
		goto cleanup;
	    }
	    o_packet++;

	    if ( bshow ) {
		gui_show_state( (gdouble)o_read/(gdouble)in_size, NULL, gui_context );
		bshow = FALSE;
	    }
	}

    }

    if ( bresult ) {
	/* calc duration */
	while ( url_ftell(iF_context->pFormatCtx->pb) <= o_read ) {
	    if ( avpacket.stream_index < 10 ) {
		duration[avpacket.stream_index] += avpacket.duration;
		av_free_packet( &avpacket );
		avpacket.stream_index = 10000;
	    }
	    if ( av_read_frame(iF_context->pFormatCtx,&avpacket) < 0 )
		break;
	}
	
	for( n = 0; 
	     n < iF_context->pFormatCtx->nb_streams && n < 10; 
	     n++ ) {
	    guint64 d = duration[n] * 1000 
		* iF_context->pFormatCtx->streams[n]->time_base.num
		/ iF_context->pFormatCtx->streams[n]->time_base.den;
	    if ( d > durationms )
		durationms = d;
	}
	iF_context->durationms = durationms;
    }

    /* End */
cleanup:
    if ( out ) {
	if ( bresult && fflush( out ) ) {
	    g_set_error( err,
			 G_FILE_ERROR,
			 g_file_error_from_errno (errno),
			 _("File can't be written correctly: %s: %s"),
			 dstpath,
			 g_strerror( errno ) );
	    bresult = FALSE;
	}
	fclose( out );
	if ( bresult ) {
	    /* Fix first null packet at first rec.ts */
	    generate_null_packets( iF_context, o_packet, durationms,
				   &null_h_packet, NULL, NULL );
	    gui_show_state( -1.0, _("Fixing rec.ts ..."), gui_context );
	    if ( ( out = fopen( dstpath, "r+b" ) ) == NULL
		 || fwrite( null_h_packet, 188, 1, out ) != 1
		 || fflush( out ) ) {
		g_set_error( err,
			     G_FILE_ERROR,
			     g_file_error_from_errno(errno),
			     _("File can't be written correctly: %s: %s"),
			     dstpath,
			     g_strerror( errno ) );
		bresult = FALSE;
	    }
	    if ( out )
		fclose( out );
	}
    }
    if ( bresult )
	iF_context->packets = o_packet;

    if ( packet )
	g_free( packet );
    if ( pat_packet )
	g_free( pat_packet );
    if ( pmt_packet )
	g_free( pmt_packet );
    if ( null_h_packet )
	g_free( null_h_packet );
    if ( null_i_packet )
	g_free( null_i_packet );
    if ( null_e_packet )
	g_free( null_e_packet );
    ifile_close( iF_context );
    fclose( in );
    g_free( dstpath );

    if ( bresult )
	gui_show_state( 1.0, NULL, gui_context );
    return bresult;
}

/**
 * create new meta.dat from ifile
 *
 * Note: you have to call rec_create_ts before to set up iF_context!
 *
 * @param dstdir The folder where to create the meta.dat.
 * @param iF_context pointer to the ifile context of the source
 * @param gui_context a gui context to report
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return TRUE on success, FALSE on error
 */
gboolean rec_create_meta(const gchar *dstdir,
			 IFileContext *iF_context,
			 GuiContext *gui_context,
			 GError **err) {
    gchar *dstpath, *out_name;
    FILE *out;
    guint8 *buffer = g_malloc0( 112 );
    gboolean bresult = TRUE;
    guint32 info_packets = (iF_context->packets + 511) / 512;
    struct stat sbuf;
    struct tm *starttime;

    buffer[0] = 0x05;
    buffer[2] = 0x72;
    buffer[3] = 0x6a;
    buffer[4] = 0x07;
    
    buffer[0x0c] = 0x02;
    strncpy( buffer + 0x0d, ifile_get_title( iF_context, NULL ), 61 );

    if ( !g_stat( iF_context->filename, &sbuf ) ) {
	g_debug( _("time = %llds\n"), (long long)sbuf.st_mtime );
	g_debug( _("duration = %llds\n"), (long long)iF_context->durationms );
	sbuf.st_mtime -= iF_context->durationms / 1000;
	starttime = gmtime( &(sbuf.st_mtime) );
	g_debug( _("Year = %d\n"), starttime->tm_year );
    } else {
	time_t etime;
	g_warning( _("Using current UTC, because cannot stat %s!\n"), 
		   iF_context->filename );
	time( &etime );
	starttime = gmtime( &etime );
    }
    buffer[0x4b] = starttime->tm_hour;  /* start time hour */
    buffer[0x4c] = starttime->tm_min;   /* start time minute */
    buffer[0x4d] = starttime->tm_sec;  /* start time second */
    
    buffer[0x50] = ((int)(1900 + starttime->tm_year) & 0xff); /* start time year LB */
    buffer[0x51] = ((int)(1900 + starttime->tm_year) >> 8) & 0xff; /* start time year HB */
    buffer[0x52] = starttime->tm_mon + 1;/* start time month */
    buffer[0x53] = starttime->tm_mday;   /* start time day */
    buffer[0x54] = starttime->tm_wday;   /* start time weekday */
    
    buffer[0x58] = (info_packets) & 0xff;
    buffer[0x59] = (info_packets>>8) & 0xff;
    buffer[0x5a] = (info_packets>>16) & 0xff;
    buffer[0x5b] = (info_packets>>24) & 0xff;

    buffer[0x5c] = iF_context->durationms / ( 1000 * 3600 );
    buffer[0x5d] = ( iF_context->durationms / ( 1000 * 60 ) ) % 60;
    buffer[0x5e] = ( iF_context->durationms / 1000 ) % 60;
    
    buffer[0x60] = 0; /* TODO: It may also be Radio! */

    buffer[0x68] = iF_context->NID & 0xff;
    buffer[0x69] = ( iF_context->NID >> 8 ) & 0xff;

    buffer[0x6c] = 1; /* always use timer number 1 */

    dstpath = g_strndup( dstdir, strlen(dstdir) + 10 ); /* 8 would be enough */
    
    out_name = dstpath + strlen( dstpath );
    if ( out_name != dstpath
	 && !G_IS_DIR_SEPARATOR( out_name[-1] ) )
	*out_name++ = G_DIR_SEPARATOR;
    strcpy( out_name, "meta.dat" ); /* That's always the first file name! */

    gui_show_state( -1.0, _("Creating meta.dat ..."), gui_context );

    if ( ( out = g_fopen( dstpath, "wb" ) ) == NULL 
	 || fwrite( buffer, 112, 1, out ) != 1
	 || fflush( out ) ) {
	g_set_error( err,
		     G_FILE_ERROR,
		     g_file_error_from_errno(errno),
		     _("File can't be written correctly: %s: %s"),
		     dstpath,
		     g_strerror( errno ) );
	bresult = FALSE;
    }
    if ( out )
	fclose( out );
    g_free( dstpath );
    g_free( buffer );

    return bresult;
}

/**
 * create new rec.cp files from ifile
 *
 * Note: creating the ts will change the iF_context!
 *
 * @param dstdir The folder where to create the file.
 * @param iF_context pointer to the ifile context of the source
 * @param gui_context a gui context to report
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return TRUE on success, FALSE on error
 */
gboolean rec_create_cp(const gchar *dstdir,
		       IFileContext *iF_context,
		       GuiContext *gui_context,
		       GError **err) {
    gchar *dstpath, *out_name;
    FILE *out;
    guint8 *buffer = g_malloc0( 372 );
    gboolean bresult = TRUE;

    dstpath = g_strndup( dstdir, strlen(dstdir) + 10 ); /* 8 would be enough */
    
    out_name = dstpath + strlen( dstpath );
    if ( out_name != dstpath
	 && !G_IS_DIR_SEPARATOR( out_name[-1] ) )
	*out_name++ = G_DIR_SEPARATOR;
    strcpy( out_name, "rec.cp" ); /* That's always the first file name! */

    gui_show_state( -1.0, _("Creating rec.cp ..."), gui_context );

    if ( ( out = g_fopen( dstpath, "wb" ) ) == NULL 
	 || fwrite( buffer, 372, 1, out ) != 1
	 || fflush( out ) ) {
	g_set_error( err,
		     G_FILE_ERROR,
		     g_file_error_from_errno(errno),
		     _("File can't be written correctly: %s: %s"),
		     dstpath,
		     g_strerror( errno ) );
	bresult = FALSE;
    }
    if ( out )
	fclose( out );
    g_free( dstpath );
    g_free( buffer );

    return bresult;
}

/**
 * create new rec.bm files from ifile
 *
 * Note: creating the ts will change the iF_context!
 *
 * @param dstdir The folder where to create the file.
 * @param iF_context pointer to the ifile context of the source
 * @param gui_context a gui context to report
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return TRUE on success, FALSE on error
 */
gboolean rec_create_bm(const gchar *dstdir,
		       IFileContext *iF_context,
		       GuiContext *gui_context,
		       GError **err) {
    gchar *dstpath, *out_name;
    FILE *out;
    guint8 *buffer = g_malloc0( 192 );
    gboolean bresult = TRUE;

    buffer[0] = 0xff;
    buffer[1] = 0xff;
    buffer[2] = 0xff;
    buffer[3] = 0xff;
    buffer[4] = 0x01;

    dstpath = g_strndup( dstdir, strlen(dstdir) + 10 ); /* 8 would be enough */
    
    out_name = dstpath + strlen( dstpath );
    if ( out_name != dstpath
	 && !G_IS_DIR_SEPARATOR( out_name[-1] ) )
	*out_name++ = G_DIR_SEPARATOR;
    strcpy( out_name, "rec.bm" ); /* That's always the first file name! */

    gui_show_state( -1.0, _("Creating rec.bm ..."), gui_context );

    if ( ( out = g_fopen( dstpath, "wb" ) ) == NULL 
	 || fwrite( buffer, 192, 1, out ) != 1
	 || fflush( out ) ) {
	g_set_error( err,
		     G_FILE_ERROR,
		     g_file_error_from_errno(errno),
		     _("File can't be written correctly: %s: %s"),
		     dstpath,
		     g_strerror( errno ) );
	bresult = FALSE;
    }
    if ( out )
	fclose( out );
    g_free( dstpath );
    g_free( buffer );

    return bresult;
}
