/*
 * ifile.h - the ifile for createrec-gtk
 * Copyright (c) 2009 Markus Kohm
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef __ifile_h
#define __ifile_h

#include <glib.h>

#define AV_STREAM_ERROR g_quark_from_static_string( "CREATEREC::AV::ERROR" )

typedef struct _ifile_context IFileContext;

/**
 * Convert to mpegts if necessary
 *
 * @param tmppath the folder to be used for the conversation destination
 * @param iF_context the file context
 * @param err Will be set on error, to describe the error, if not NULL.
 * @result TRUE if the conversation was successfull or not needed, 
 *         FALSE on error.
 */
gboolean ifile_convert( const gchar *tmppath,
			IFileContext *iF_context, GError **err );

/**
 * Get the current file name from input file context.
 *
 * @param iF_context the input file context.
 * @result either the file name from the given input file context or NULL
 */
const gchar *ifile_get_filename( IFileContext *iF_context );

/**
 * Set the current file name and do all needed initializations for input.
 *
 * @param filename the new filename to be used for input.
 * @param iF_context the input file context.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @result < 0 if the file name cannot be used for input, 
 *         = 0 if the file name can be used for direct input, 
 *         > 0 if the file name can be used for input after conversation 
 *             of the input format.
 */
gint ifile_set_filename( const gchar *filename, 
			 IFileContext *iF_context, GError **err );

/**
 * get the title of the file
 *
 * @param iF_context the input file context.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return title or NULL
 */
const gchar *ifile_get_title( IFileContext *iF_context, GError **err );

/**
 * set the title of the file
 *
 * @param title the new title of the file
 * @param iF_context the input file context.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return TRUE on success, FALSE on error
 */
gboolean ifile_set_title( const gchar *title, 
			  IFileContext *iF_context, GError **err );

/**
 * Do all initialization for unknown input files.
 *
 * Note, that argc and argv are input and output parameters.  After
 * the initialization all gui parameters are removed from the array.
 *
 * @param argc pointer to number of arguments.
 * @param argv array of arguments.
 * @param iF_context pointer to a ifile context.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return TRUE if ok, FALSE on error.
 */
gboolean ifile_init(int *argc, char ***argv, 
		    IFileContext **iF_context, GError **err);

/**
 * Open the file from the file context (if not already open)
 *
 * Note: You also have to close the file!
 *
 * @param iF_context the context of the file, that should be opened
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return TRUE if the file was opened, FALSE on error
 */
gboolean ifile_open(IFileContext *iF_context, GError **err);

/**
 * Close the file from the file context (if open)
 *
 * Note: The file context is still valid after closing. You have to use
 *       ifile_exit zu destroy it.
 *
 * @param iF_context the context of the file, that should be closed
 */
void ifile_close(IFileContext *iF_context);

/**
 * Ask for the format of the file.
 *
 * Note: If the file was currently not opened, it will be opened temporary 
 *       only and only if neccessary.
 *
 * @param iF_context the context of the file.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return name string of the file format or NULL on error.
 */
const gchar *ifile_get_format(IFileContext *iF_context, GError **err);

/**
 * Ask for the number of video streams at the file.
 *
 * Note: if the file was currently not opened, it will be.
 *
 * @param iF_context the context of the file.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return number of video streams at the file or -1 on error.
 */
gint ifile_get_video_stream_count(IFileContext *iF_context, GError **err);

/**
 * Ask for the codec of a video stream.
 *
 * Note: if the file was currently not opened, it will be.
 *
 * @param iF_context the context of the file.
 * @param stream_number The number of the video stream.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return name string of the codec or NULL on error or if not enough
 *              streams where found.
 */
const gchar *ifile_get_video_codec_name(IFileContext *iF_context, 
					guint stream_number,
					GError **err);
enum CodecID ifile_get_video_codec_id(IFileContext *iF_context, 
				      guint stream_number,
				      GError **err);

/**
 * Ask for the number of audio streams at the file.
 *
 * Note: if the file was currently not opened, it will be.
 *
 * @param iF_context the context of the file.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return number of audio streams at the file or -1 on error.
 */
gint ifile_get_audio_stream_count(IFileContext *iF_context, GError **err);

/**
 * Ask for the codec of a audio stream.
 *
 * Note: if the file was currently not opened, it will be.
 *
 * @param iF_context the context of the file.
 * @param stream_number The number of the audio stream.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return name string of the codec or NULL on error or if not enough
 *              streams where found.
 */
const gchar *ifile_get_audio_codec_name(IFileContext *iF_context,
					guint stream_number,
					GError **err);
enum CodecID ifile_get_audio_codec_id(IFileContext *iF_context, 
				      guint stream_number,
				      GError **err);

/**
 * Ask for the language of an audio stream.
 *
 * Note: if the file was currently not opened, it will be.
 *
 * @param iF_context the context of the file.
 * @param stream_number The number of the audio stream.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return name string of the language or NULL on error or if not enough
 *              streams where found.
 */
const gchar *ifile_get_audio_lang(IFileContext *iF_context, 
				  guint stream_number,
				  GError **err);

/**
 * Ask for the number of subtitle streams at the file.
 *
 * Note: if the file was currently not opened, it will be.
 *
 * @param iF_context the context of the file.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return number of subtitle streams at the file or -1 on error.
 */
gint ifile_get_subtitle_stream_count(IFileContext *iF_context,
				     GError **err);

/**
 * Ask for the language of an subtitle stream.
 *
 * Note: if the file was currently not opened, it will be.
 *
 * @param iF_context the context of the file.
 * @param stream_number The number of the subtitle stream.
 * @param err Will be set on error, to describe the error, if not NULL.
 * @return name string of the language or NULL on error or if not enough
 *              streams where found.
 */
const gchar *ifile_get_subtitle_lang(IFileContext *iF_context,
				     guint stream_number,
				     GError **err);

/**
 * Exit the ifile
 *
 * Note, that iF_context is still valid after the call. Use ifile_exit() to
 * additionaly free iF_context.
 *
 * @param iF_context pointer to a ifile context.
 * @return TRUE if exit, FALSE if not
 */
gboolean ifile_exit0(IFileContext *iF_context);

/**
 * Exit the ifile
 *
 * Note, that iF_context is not longer valid after this call!
 *
 * @param iF_context pointer to a ifile context.
 * @return TRUE if exit, FALSE if not
 */
gboolean ifile_exit(IFileContext *iF_context);

#endif /* __ifile_h */
