/*
 * createrec-gtk
 * Copyright (c) 2009 Markus Kohm
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define __USE_FILE_OFFSET64

#include <errno.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <glib/gprintf.h>
#include <glib/gi18n.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include "gui.h"
#include "ifile.h"
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>

#define CMDLINE_MAX 1000 /* max. length of the ffmpeg command line without file
			  * names */

static IFileContext *convert_input( IFileContext *iF_context,
				    const gchar *dstdir,
				    GuiContext *gui_context, 
				    GError **err ) {
    const gchar *codec;
    gchar *ffmpeg_par, *cmdline, *dstfilename, *s_h, *srctitle;
    const gchar *srcfilename;
    guint n_streams, n;
    guint32 n_r, n_cnt;

    g_assert( iF_context && dstdir && gui_context );
    
    gui_show_state( 0.0, _("Converting into MPEG TS ..."), gui_context );

    dstfilename = g_strndup( dstdir, strlen( dstdir ) + 32 );
    s_h = dstfilename + strlen( dstdir );
    if ( strlen( dstdir ) > 0 && !G_IS_DIR_SEPARATOR( s_h[-1] ) )
	*s_h++ = G_DIR_SEPARATOR;
    s_h = g_stpcpy( s_h, "createrec-" );
    for ( n_cnt = 0, n_r = g_random_int();
	  n_cnt < 1000; /* if after 1000 trys I found no usable tmp file stop */
	  n_cnt++, n_r = g_random_int() ) {
	int desc;
	g_sprintf( s_h, "%08x.ts", n_r );
	if ( g_access( dstfilename, F_OK ) 
	     && ( ( desc = g_creat( dstfilename, S_IRUSR|S_IWUSR ) ) != -1 ) ) {
	    close( desc );
	    break;
	} else {
	    g_warning( _("Cannot create file: %s: %s!\n"), 
		       dstfilename, g_strerror( errno ) );
	}
    }
    if ( n_cnt == 1000 )
	g_error( _("Cannot find a valid temporary file name!\n") );

    srcfilename = ifile_get_filename( iF_context );
    ffmpeg_par = g_malloc0( CMDLINE_MAX 
			    + strlen( srcfilename )
			    + strlen( dstfilename ) );
    cmdline = g_stpcpy( ffmpeg_par, "ffmpeg " ); /* NOTE: we need ffmpeg!!! */
    cmdline = g_stpcpy( cmdline, "-i " ); /* input file */
    s_h = g_shell_quote( srcfilename );
    cmdline = g_stpcpy( cmdline, s_h );
    g_free( s_h );
    *cmdline++ = ' ';

    cmdline = g_stpcpy( cmdline, "-y " ); /* overwrite output file */

    if ( !ifile_open( iF_context, err ) ) { /* open the file for analyzation */
	g_free( dstfilename );
	g_free( ffmpeg_par );
	return NULL;
    }

    n_streams = ifile_get_video_stream_count( iF_context, err );
    for ( n = 0; n < n_streams; n++ ) {
	enum CodecID codec;
	codec = ifile_get_video_codec_id( iF_context, n, err );
	if ( codec == CODEC_ID_MPEG2VIDEO
	    || codec == CODEC_ID_MPEG2TS )
	    cmdline = g_stpcpy( cmdline, "-vcodec copy " );
	else
	    cmdline = g_stpcpy( cmdline, "-vcodec mpeg2video " );
	if ( n != 0 ) /* NOTE: I do not know, if such streams may be handled */
	    cmdline = g_stpcpy( cmdline, "-newvideo " );
    }

    n_streams = ifile_get_audio_stream_count( iF_context, err );
    for ( n = 0; n < n_streams; n++ ) {
	const gchar *lang;
	enum CodecID codec;
	codec = ifile_get_audio_codec_id( iF_context, n, err );
	if ( ( codec >= CODEC_ID_PCM_S16LE 
	       && codec < CODEC_ID_ADPCM_IMA_QT )
	     || codec == CODEC_ID_MP1
	     || codec == CODEC_ID_MP2
	     || codec == CODEC_ID_MP3
	     || codec == CODEC_ID_AAC
	     || codec == CODEC_ID_AC3 )
	    cmdline = g_stpcpy( cmdline, "-acodec copy " );
	else
	    cmdline = g_stpcpy( cmdline, "-acodec ac3 " );
	lang = ifile_get_audio_lang( iF_context, n, err );
	if ( lang ) {
	    cmdline = g_stpcpy( cmdline, "-alang " );
	    cmdline = g_stpcpy( cmdline, lang );
	    *cmdline++ = ' ';
	}
	if ( n != 0 )
	    cmdline = g_stpcpy( cmdline, "-newaudio " );
    }

    n_streams = ifile_get_subtitle_stream_count( iF_context, err );
    for ( n = 0; n < n_streams; n++ ) {
	const gchar *lang;
	cmdline = g_stpcpy( cmdline, "-scodec copy " );
	lang = ifile_get_subtitle_lang( iF_context, n, err );
	if ( lang ) {
	    cmdline = g_stpcpy( cmdline, "-slang " );
	    cmdline = g_stpcpy( cmdline, lang );
	    *cmdline++ = ' ';
	}
	if ( n != 0 )
	    cmdline = g_stpcpy( cmdline, "-newsubtitle " );
    }

    cmdline = g_stpcpy( cmdline, "-f mpegts " ); /* force output format */

    s_h = g_shell_quote( dstfilename );
    cmdline = g_stpcpy( cmdline, s_h );
    g_free( s_h );

    srctitle = g_strdup( ifile_get_title( iF_context, NULL ) );
    ifile_close( iF_context ); /* close the file */
    
    if ( !gui_spawn_sync( ffmpeg_par, gui_context, err ) ) {
	g_free( srctitle );
	g_free( dstfilename );
	return NULL;
    }

    if ( !ifile_init( NULL, NULL, &iF_context, err ) ) {
	iF_context = NULL;
    } else if ( ifile_set_filename( dstfilename, iF_context, err ) 
	|| !ifile_set_title( srctitle, iF_context, err ) ) {
	ifile_exit( iF_context );
	iF_context = NULL;
    } else {
	struct stat sbuf;
	if ( !g_stat( srcfilename, &sbuf ) )
	    utimes( dstfilename, &(sbuf.st_atime) );
    }

    g_free( srctitle );
    g_free( dstfilename );

    return iF_context;
}

gboolean createrec_do( const gchar *dirname, 
		       guint number, gboolean searchfirstfree,
		       IFileContext *iF_context, const gchar *tmpdir,
		       GuiContext *gui_context, 
		       GError **err ) {
    IFileContext *iF;
    gboolean bresult;
    gchar *dstpath = g_strndup( dirname, strlen(dirname)+20 ); /* 18 would be enough*/
    gchar *dstfile = dstpath + strlen( dirname );
    gint n;

    gui_show_state( 0.0, _("Creating record ..."), gui_context );

    if ( dstfile != dstpath && !G_IS_DIR_SEPARATOR( dstfile[-1] ) )
	*dstfile++ = G_DIR_SEPARATOR;

    g_sprintf( dstfile, "rec_%04u", number );
    errno = 0;
    bresult = FALSE;
    if ( g_access( dstpath, F_OK ) 
	 && errno == ENOENT
	 && !g_mkdir_with_parents( dstpath, ACCESSPERMS ) ) {
	bresult = TRUE;
    } else if ( searchfirstfree ) {
	for ( n = number+1; !bresult && n < 10000; n++ ) {
	    g_sprintf( dstfile, "rec_%04d", n );
	    if ( g_access( dstpath, F_OK )
		 && errno == ENOENT
		 && !g_mkdir_with_parents( dstpath, ACCESSPERMS ) ) {
		bresult = TRUE;
	    }
	}
	if ( !bresult && number > 0 )
	    for ( n = 0; !bresult && n <= number-1; n++ ) {
		g_sprintf( dstfile, "rec_%04d", n );
		if ( g_access( dstpath, F_OK )
		     && errno == ENOENT
		     && !g_mkdir_with_parents( dstpath, ACCESSPERMS ) ) {
		    bresult = TRUE;
		}
	    }
    }
    if ( !bresult ) {
	g_set_error( err,
		     g_quark_from_static_string( "CREATEREC::ERROR" ),
		     AVERROR( errno ? errno : ENOENT ),
		     _("Cannot create record directory at %s: %s"),
		     dirname,
		     errno ? g_strerror( errno ) : _("directory does alreay exist") );
	g_free( dstpath );
	return FALSE;
    }
    dstfile = dstfile + strlen( dstfile );

    if ( g_strcmp0( ifile_get_format( iF_context, err ), "mpegts" ) ) {
	/* We need to convert the input */
	iF = convert_input( iF_context, tmpdir, gui_context, err );
	if ( !iF )
	    bresult = FALSE;
    } else {
	iF = iF_context;
    }

    if ( bresult )
	bresult = rec_create_ts( dstpath, iF, gui_context, err );

    if ( bresult )
	bresult = rec_create_meta( dstpath, iF, gui_context, err );

    if ( bresult )
	bresult = rec_create_cp( dstpath, iF, gui_context, err );

    if ( bresult )
	bresult = rec_create_bm( dstpath, iF, gui_context, err );

    ifile_close( iF_context );

    if ( iF != iF_context ) {
	const gchar *filename = ifile_get_filename( iF );
	ifile_close( iF );
	if ( filename ) {
	    gui_show_state( -1.0, _("Deleting temporary files ..."), gui_context );
	    if ( g_unlink( filename ) < 0 ) {
		g_set_error( err,
			     G_FILE_ERROR,
			     AVERROR( errno ),
			     _("Cannot delete temporary file: %s: %s"),
			     filename, g_strerror( errno ) );
	    }
	}
	ifile_exit( iF );
    }

    if ( !bresult ) {
	*dstfile = '\0';
	g_remove( dstpath );
    } else {
	gui_show_state( -1.0, _("All done!"), gui_context );
    }

    g_free( dstpath );
    return bresult;
}

int main( int   argc,
          char *argv[] ) 
{
    GuiContext *context;

    setlocale( LC_ALL, "" );
    textdomain ( PACKAGE );
    bindtextdomain( PACKAGE, LOCALEDIR );
 
    gui_init (&argc, &argv, &context);

    gui_do (context);

    return 0;
}
